//    Copyright (C) 2009 Dirk Vanden Boer <dirk.vdb@gmail.com>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "flacdecoder.h"

#include <stdexcept>

#include "AudioRenderer/audioframe.h"
#include "Utils/scopedlock.h"
#include "Utils/log.h"

using namespace std;
using namespace utils;

namespace Gejengel
{

FlacDecoder::FlacDecoder(const std::string& filepath)
: AudioDecoder(filepath)
, m_BytesPerFrame(0)
, m_NumSamples(0)
{
    initialize();
}

FlacDecoder::~FlacDecoder()
{
    destroy();
}

void FlacDecoder::destroy()
{
    m_BytesPerFrame = 0;
    finish();
}

void FlacDecoder::initialize()
{
    set_md5_checking(true);
    
    FLAC__StreamDecoderInitStatus initStatus = init(m_Filepath);
    if (initStatus != FLAC__STREAM_DECODER_INIT_STATUS_OK)
    {
        throw logic_error(FLAC__StreamDecoderInitStatusString[initStatus]);
    }

    if (!process_until_end_of_metadata())
    {
        throw logic_error("Failed to parse flac metadata");
    }
}

AudioFormat FlacDecoder::getAudioFormat()
{
    return m_Format;
}

double FlacDecoder::getProgress()
{
    double duration = getDuration();
    return duration == 0.0 ? 0.0 : m_AudioClock / duration;
}

double FlacDecoder::getDuration()
{
    if (m_Format.rate == 0)
    {
        return 0.0;
    }

    return static_cast<double>(m_NumSamples / m_Format.rate);
}

int32_t FlacDecoder::getFrameSize()
{
    return m_BytesPerFrame;
}

void FlacDecoder::seekAbsolute(double time)
{
    double duration = getDuration();
    if (duration == 0.0 || m_NumSamples == 0)
    {
        return;
    }
    
    FLAC__uint64 sampleNr = static_cast<FLAC__uint64>((time / duration) * m_NumSamples);
    if (!seek_absolute(sampleNr))
    {
        Log::warn("Seek failed");
    }
}

void FlacDecoder::seekRelative(double offset)
{
    seekAbsolute(m_AudioClock + offset);
}

bool FlacDecoder::decodeAudioFrame(AudioFrame& frame)
{
    if (!process_single())
    {
        Log::error("Flac decode error");
        return false;
    }

    if (get_state() == FLAC__STREAM_DECODER_END_OF_STREAM)
    {
        Log::debug("End of stream");
        return false;
    }

    frame.setFrameData(&m_AudioBuffer[0]);
    frame.setDataSize(m_AudioBuffer.size());
    frame.setPts(m_AudioClock);
    return true;
}

FLAC__StreamDecoderWriteStatus FlacDecoder::write_callback(const FLAC__Frame* pFrame, const FLAC__int32* const pBuffer[])
{
    assert(pFrame);
    assert(pFrame->header.bits_per_sample == 16);
    
    uint32_t frameSize = pFrame->header.blocksize * pFrame->header.channels * (pFrame->header.bits_per_sample / 8);
    m_AudioBuffer.resize(frameSize);

    uint16_t* pFrameData = reinterpret_cast<uint16_t*>(&m_AudioBuffer[0]);

    for (size_t i = 0; i < pFrame->header.blocksize; ++i)
    {
        *pFrameData++ = pBuffer[0][i]; //left channel
        *pFrameData++ = pBuffer[1][i]; //right channel
    }

    if (pFrame->header.number_type == FLAC__FRAME_NUMBER_TYPE_FRAME_NUMBER)
    {
        Log::debug("TODO");
    }
    else
    {
        assert(pFrame->header.number_type == FLAC__FRAME_NUMBER_TYPE_SAMPLE_NUMBER);
        m_AudioClock = pFrame->header.number.frame_number / m_Format.rate;
    }

    return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

void FlacDecoder::metadata_callback(const FLAC__StreamMetadata* pMetadata)
{
    assert(pMetadata);
    if (pMetadata->type == FLAC__METADATA_TYPE_STREAMINFO)
    {
        m_Format.bits           = pMetadata->data.stream_info.bits_per_sample;
        m_Format.rate           = pMetadata->data.stream_info.sample_rate;
        m_Format.numChannels    = pMetadata->data.stream_info.channels;
        m_NumSamples            = pMetadata->data.stream_info.total_samples;

        Log::debug("Flac Audio format: bits (", m_Format.bits, ") rate (", m_Format.rate, ") numChannels (", m_Format.numChannels, ")");
    }
}

void FlacDecoder::error_callback(FLAC__StreamDecoderErrorStatus status)
{
    Log::error("FlacDecoder:",  FLAC__StreamDecoderErrorStatusString[status]);
}

}
