//    Copyright (C) 2009 Dirk Vanden Boer <dirk.vdb@gmail.com>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "gejengel.h"

#include "Utils/log.h"
#include "Utils/scopedlock.h"
#include "Utils/timeoperations.h"
#include "Utils/numericoperations.h"
#include "MusicLibrary/subscribers.h"
#include "playbackengine.h"
#include "playbackenginefactory.h"

#include <glibmm/i18n.h>

using namespace std;

namespace Gejengel
{

GejengelCore::GejengelCore()
: m_Library(m_Settings)
, m_PlayQueue(m_Library)
, m_pPlayback(PlaybackEngineFactory::create(m_Settings.get("PlaybackEngine"), *this))
, m_PluginMgr(*this)
, m_MainWindow(*this)
, m_CurrentProgress(0)
, m_FetchStatus(FetchStarted)
{
    m_Library.addLibrarySubscriber(m_MainWindow);

    dispatchProgress.connect(sigc::mem_fun(this, &GejengelCore::sendProgress));
    dispatchVolumeChanged.connect(sigc::mem_fun(this, &GejengelCore::sendVolumeChange));
    dispatchStop.connect(sigc::mem_fun(&m_PluginMgr, &PluginManager::sendStop));
    dispatchNewTrackStarted.connect(sigc::mem_fun(this, &GejengelCore::newTrackStarted));
    dispatchGetNextTrack.connect(sigc::mem_fun(this, &GejengelCore::sendGetNextTrack));

    m_PlayQueue.loadQueue();
    m_PlayQueue.subscribe(*this);

    loadSettings();
}

GejengelCore::~GejengelCore()
{
    m_PlayQueue.unsubscribe(*this);
    saveSettings();
    delete m_pPlayback;
}

void GejengelCore::loadSettings()
{
    m_pPlayback->setVolume(m_Settings.getAsInt("PlaybackVolume", 100));
}

void GejengelCore::saveSettings()
{
    m_Settings.set("PlaybackVolume", m_pPlayback->getVolume());

    if (m_Settings.getAsBool("SaveQueueOnExit", false))
    {
        m_PlayQueue.saveQueue();
    }
}

void GejengelCore::run()
{
    m_MainWindow.run();
}

void GejengelCore::play()
{
    if (!m_pPlayback->isPlaying())
    {
        try
        {
            m_pPlayback->play();
        }
        catch(exception& e)
        {
            Log::error(string("Failed to start playback: ") + e.what());
        }
    }
}

void GejengelCore::pause()
{
    if (m_pPlayback->isPlaying())
    {
        try
        {
            m_pPlayback->pause();
            m_PluginMgr.sendPause();
        }
        catch(exception& e)
        {
            Log::error(string("Failed to pause playback: ") + e.what());
        }
    }
}

void GejengelCore::resume()
{
    if (!m_pPlayback->isPlaying())
    {
        try
        {
            m_pPlayback->resume();
            m_PluginMgr.sendResume();
        }
        catch(exception& e)
        {
            Log::error(string("Failed to resume playback: ") + e.what());
        }
    }
}

void GejengelCore::prev()
{
    m_pPlayback->prev();
}

void GejengelCore::next()
{
    m_pPlayback->next();
}

void GejengelCore::stop()
{
    try
    {
        m_CurrentTrack = Track();
        m_pPlayback->stop();
        m_PluginMgr.sendStop();
    }
    catch(exception& e)
    {
        Log::error(string("Failed to stop playback: ") + e.what());
    }
}

PlaybackState GejengelCore::getPlaybackState()
{
    return m_pPlayback->getState();
}

void GejengelCore::seek(double seconds)
{
    try
    {
        m_pPlayback->seek(seconds);
    }
    catch(exception& e)
    {
        Log::error(string("Failed to seek: ") + e.what());
    }
}

double GejengelCore::getTrackPosition()
{
    return m_pPlayback->getCurrentTime();
}

void GejengelCore::getTrack(Track& track)
{
    track = m_pPlayback->getTrack();
}

void GejengelCore::setVolume(int32_t volume)
{
    m_pPlayback->setVolume(volume);
}

int32_t GejengelCore::getVolume()
{
    return m_pPlayback->getVolume();
}

Settings& GejengelCore::getSettings()
{
    return m_Settings;
}

PluginManager& GejengelCore::getPluginManager()
{
    return m_PluginMgr;
}

PlayQueue& GejengelCore::getPlayQueue()
{
    return m_PlayQueue;
}

MusicLibrary& GejengelCore::getMusicLibrary()
{
    return m_Library;
}

IAlbumArtProvider& GejengelCore::getAlbumArtProvider()
{
    return *this;
}

void GejengelCore::onTrackQueued(uint32_t index)
{
    play();
}

void GejengelCore::quitApplication()
{
    m_MainWindow.quit();
}

void GejengelCore::showHideWindow()
{
    m_MainWindow.showHideWindow();
}

bool GejengelCore::getNextTrack(Track& track)
{
    m_FetchStatus = FetchStarted;
    dispatchGetNextTrack();

    while (m_FetchStatus == FetchStarted)
    {
        TimeOperations::sleepMs(10);
    }

    if (m_FetchStatus == FetchSuccess)
    {
        track = m_CurrentTrack;
        return true;
    }

    dispatchStop();
    return false;
}

void GejengelCore::newTrackStarted()
{
    m_CurrentTrack = m_pPlayback->getTrack();
    m_PluginMgr.sendPlay(m_CurrentTrack);
}

void GejengelCore::sendGetNextTrack()
{
    m_FetchStatus = m_PlayQueue.getNextTrack(m_CurrentTrack) ? FetchSuccess : FetchFailed;
}

void GejengelCore::getAlbum(Album& album)
{
    if (m_CurrentTrack.albumId == 0)
    {
        return;
    }

    m_Library.getAlbum(m_CurrentTrack.albumId, album);
    getAlbumArt(album.id, album.coverData);
}

bool GejengelCore::getAlbumArt(uint32_t albumId, std::vector<uint8_t>& data)
{
    return m_Library.getAlbumArt(albumId, data);
}

void GejengelCore::sendProgress()
{
    m_PluginMgr.sendProgress(static_cast<int32_t>(m_pPlayback->getCurrentTime()));
}

void GejengelCore::sendVolumeChange()
{
    m_PluginMgr.sendVolumeChanged(m_pPlayback->getVolume());
}

}
