//    Copyright (C) 2009 Dirk Vanden Boer <dirk.vdb@gmail.com>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "playback.h"

#include <iomanip>
#include <cmath>
#include <cassert>
#include <cstring>

#include "AudioDecoder/audiodecoder.h"
#include "AudioDecoder/audiodecoderfactory.h"
#include "AudioRenderer/audiorenderer.h"
#include "AudioRenderer/audiorendererfactory.h"
#include "MusicLibrary/track.h"
#include "MusicLibrary/album.h"
#include "Utils/timeoperations.h"
#include "Utils/numericoperations.h"
#include "Utils/scopedlock.h"
#include "Utils/log.h"
#include "Utils/trace.h"
#include "gejengel.h"
#include "gejengelplugin.h"
#include "musiclibrary.h"
#include "settings.h"

using namespace std;

namespace Gejengel
{

Playback::Playback(GejengelCore& core)
: PlaybackEngine(core)
, m_pAudioDecoder(NULL)
, m_pAudioRenderer(AudioRendererFactory::create(core.getSettings().get("AudioBackend"), *this))
, m_PlaybackThread(Playback::playbackThread, this)
, m_Destroy(false)
, m_Stop(false)
, m_NewTrackStarted(false)
, m_State(Stopped)
, m_SkipTrack(false)
, m_SeekOccured(false)
, m_CurrentPts(0)
{
    Utils::trace("Create Playback");
    m_PlaybackThread.start();

#ifdef DUMP_TO_WAVE
    Log::critical("Dumping to wave is enabled, we are debugging right?");
    writeWaveHeader();
#endif
}


Playback::~Playback()
{
    stop();
    m_Destroy = true;

    #ifdef DUMP_TO_WAVE
    updateWaveHeaderSize();
    #endif

    {
        utils::ScopedLock lock(m_PlaybackMutex);
        m_PlaybackCondition.broadcast();
    }

    Log::debug("Waiting for playback thread");
    m_PlaybackThread.join();
    Log::debug("Done waiting");

    delete m_pAudioRenderer;
}

bool Playback::startNewTrack()
{
    if (!m_Core.getNextTrack(m_CurrentTrack))
    {
        stopPlayback(true);
        return false;
    }

    m_CurrentPts = 0.0;

    {
        utils::ScopedLock lock(m_DecodeMutex);
        try
        {
            delete m_pAudioDecoder;
            m_pAudioDecoder = AudioDecoderFactory::create(m_CurrentTrack.filepath);
        }
        catch (logic_error& e)
        {
            Log::error("Failed to play audio file: ", e.what());
            m_pAudioDecoder = NULL;
            return startNewTrack();
        }
    }

    uint32_t duration = static_cast<uint32_t>(m_pAudioDecoder->getDuration());
    if (duration != 0)
    {
        m_CurrentTrack.durationInSec = duration;
    }
    m_Core.dispatchNewTrackStarted();
    m_NewTrackStarted = true;

    AudioFormat format = m_pAudioDecoder->getAudioFormat();
    m_pAudioRenderer->setFormat(format);

    return true;
}

void Playback::playback()
{
    if (!startNewTrack())
    {
        return;
    }

    m_pAudioRenderer->play();

    while (!m_Stop)
    {
        while (!m_Stop && !isPaused() && rendererHasSpace())
        {
            utils::ScopedLock lock(m_PlaybackMutex);
            if (m_SkipTrack)
            {
                if (!startNewTrack())
                {
                    return;
                }
                m_pAudioRenderer->play();
                m_SkipTrack = false;
            }

            if (isPaused())
            {
                break;
            }

            bool frameDecoded = false;
            AudioFrame audioFrame;
            if (m_pAudioDecoder)
            {
                utils::ScopedLock lock(m_DecodeMutex);
                frameDecoded = m_pAudioDecoder->decodeAudioFrame(audioFrame);
            }

            if (frameDecoded)
            {
                m_pAudioRenderer->queueFrame(audioFrame);
                #ifdef DUMP_TO_WAVE
                dumpToWav(audioFrame);
                #endif
            }
            else if (!startNewTrack())
            {
                return;
            }

            sendProgressIfNeeded();
        }

        m_pAudioRenderer->flushBuffers();

        if (m_State == Playing && !m_pAudioRenderer->isPlaying())
        {
            Log::debug("Kick renderer");
            utils::ScopedLock lock(m_PlaybackMutex);
            m_pAudioRenderer->play();
        }

        TimeOperations::sleepMs(50);
    }
}

bool Playback::rendererHasSpace()
{
    if (m_pAudioDecoder)
    {
        utils::ScopedLock lock(m_PlaybackMutex);
        return m_pAudioRenderer->hasBufferSpace(m_pAudioDecoder->getFrameSize());
    }

    return false;
}

void Playback::sendProgressIfNeeded()
{
    double pts = m_pAudioRenderer->getCurrentPts();

    if (pts <= 1 && m_NewTrackStarted)
    {
        m_NewTrackStarted = false;
    }

    if (static_cast<int>(pts) == 0 || m_NewTrackStarted)
    {
        //avoid sending false progress after a seek, or track that has just been started
        return;
    }

    if ((static_cast<int>(pts) != static_cast<int>(m_CurrentPts)))
    {
        m_CurrentPts = pts;
        m_Core.dispatchProgress();
    }
}

void Playback::play()
{
    if (m_State == Playing || m_State == Paused)
    {
        return;
    }

    utils::ScopedLock lock(m_PlaybackMutex);
    m_PlaybackCondition.broadcast();
    m_State = Playing;
}

void Playback::pause()
{
    utils::ScopedLock lock(m_PlaybackMutex);
    m_pAudioRenderer->pause();
    m_State = Paused;
}

void Playback::resume()
{
    if (isPaused())
    {
        utils::ScopedLock lock(m_PlaybackMutex);
        m_SeekOccured ? m_pAudioRenderer->play() : m_pAudioRenderer->resume();
        m_State = Playing;
    }
    else
    {
        Log::debug("Not in paused state, ignore resume");
    }
}

void Playback::stop()
{
    stopPlayback(false);
}

void Playback::stopPlayback(bool drain)
{
    m_CurrentTrack = Track();
    m_CurrentPts = 0.0;
    m_Stop = true;
    utils::ScopedLock lock(m_PlaybackMutex);
    m_pAudioRenderer->stop(drain);
    m_State = Stopped;
    m_SeekOccured = false;
    m_NewTrackStarted = false;
}

void Playback::next()
{
    if (m_SkipTrack || m_State == Stopped)
    {
        //previous skip not processed yet or we are stopped
        return;
    }

    utils::ScopedLock lock(m_PlaybackMutex);

    m_SkipTrack = true;
    m_pAudioRenderer->stop(false);
    m_pAudioRenderer->flushBuffers();
    m_State = Playing;

    m_CurrentPts = 0.0;
    m_Core.dispatchProgress();
}

void Playback::prev()
{
}

bool Playback::isPaused()
{
    return m_State == Paused;
}

bool Playback::isPlaying()
{
    return m_pAudioRenderer->isPlaying();
}

void Playback::seek(double seconds)
{
    if (!m_pAudioDecoder)
    {
        return;
    }

    utils::ScopedLock playbackLock(m_PlaybackMutex);
    m_pAudioRenderer->stop(false);
    m_pAudioRenderer->flushBuffers();

    {
        utils::ScopedLock decodeLock(m_DecodeMutex);
        m_pAudioDecoder->seekAbsolute(static_cast<double>(seconds));
    }

    if (!isPaused())
    {
        m_pAudioRenderer->play();
    }
    else
    {
        m_SeekOccured = true;
    }

    m_NewTrackStarted = false;
}

double Playback::getCurrentTime()
{
    return m_CurrentPts;
}

double Playback::getDuration()
{
    return m_pAudioDecoder ? static_cast<double>(m_pAudioDecoder->getDuration()) : 0.0;
}

PlaybackState Playback::getState()
{
    return m_State;
}

void Playback::setVolume(int32_t volume)
{
    {
        utils::ScopedLock lock(m_PlaybackMutex);
        m_pAudioRenderer->setVolume(volume);
    }
    m_Core.dispatchVolumeChanged();
}

int32_t Playback::getVolume() const
{
    return m_pAudioRenderer->getVolume();
}

const Track& Playback::getTrack() const
{
    return m_CurrentTrack;
}

void Playback::playbackLoop()
{
    while (!m_Destroy)
    {
        {
            Log::debug("Wait");
            utils::ScopedLock lock(m_PlaybackMutex);
            m_PlaybackCondition.wait(m_PlaybackMutex);
            Log::debug("Condition signaled playback= " + NumericOperations::toString(m_Stop));
        }

        if (m_Destroy)
        {
            break;
        }

        try
        {
            m_Stop = m_SkipTrack = false;
            playback();
        }
        catch (exception& e)
        {
            Log::error(string("Playback error: ") + e.what());
        }
    }
}

void* Playback::playbackThread(void* pInstance)
{
    Playback* pPlayback = reinterpret_cast<Playback*>(pInstance);
    pPlayback->playbackLoop();
    return NULL;

}

#ifdef DUMP_TO_WAVE

void Playback::writeWaveHeader()
{
    m_WaveFile.open("/tmp/dump.wav", ios_base::binary);
    if (!m_WaveFile.is_open())
    {
        return;
    }

    uint16_t bps = 16;
    uint32_t subchunk1Size = 16; //16 for PCM
    uint16_t audioFormat = 1; //PCM
    uint16_t channels = 2;
    uint32_t sampleRate = 44100;
    uint32_t byteRate = 44100 * 2 * (bps / 8);
    uint16_t blockAlign = 2 * (bps / 8);

    //don't know yet
    m_WaveBytes = 0;

    m_WaveFile.write("RIFF", 4);
    m_WaveFile.write((char*) &m_WaveBytes, 4);
    m_WaveFile.write("WAVEfmt ", 8);
    m_WaveFile.write((char*) &subchunk1Size, 4);
    m_WaveFile.write((char*) &audioFormat, 2);
    m_WaveFile.write((char*) &channels, 2);
    m_WaveFile.write((char*) &sampleRate, 4);
    m_WaveFile.write((char*) &byteRate, 4);
    m_WaveFile.write((char*) &blockAlign, 2);
    m_WaveFile.write((char*) &bps, 2);
    m_WaveFile.write("data", 4);
    m_WaveFile.write((char*) &m_WaveBytes, 4);
}

void Playback::updateWaveHeaderSize()
{
    m_WaveFile.seekp(4, ios_base::beg);
    m_WaveFile.write((char*) &m_WaveBytes + 36, 4);

    m_WaveFile.seekp(40, ios_base::beg);
    m_WaveFile.write((char*) &m_WaveBytes, 4);

    m_WaveFile.close();
}

void Playback::dumpToWav(AudioFrame& frame)
{
    assert(m_WaveFile.is_open());
    m_WaveFile.write(reinterpret_cast<char*>(frame.getFrameData()), frame.getDataSize());
    m_WaveBytes += frame.getDataSize();
}

#endif
}
