//    Copyright (C) 2009 Dirk Vanden Boer <dirk.vdb@gmail.com>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "playqueue.h"

#include "Utils/log.h"
#include "Utils/fileoperations.h"
#include "Core/gejengel.h"

#include <fstream>

using namespace std;

#ifndef WIN32
#define CONFIG_DIR     ".config/gejengel"
#else
#define CONFIG_DIR     "gejengel"
#endif
#define QUEUE_FILE     "gejengel.q"

namespace Gejengel
{

class QueueTrackSubscriber : public ITrackSubscriber
{
public:
    QueueTrackSubscriber(PlayQueue& queue, int32_t startIndex = -1)
    : m_PlayQueue(queue)
    , m_Index(startIndex)
    {}

    void onTrack(const Track& track)
    {
        m_PlayQueue.queueTrack(track, m_Index);

        if (m_Index != -1)
        {
            ++m_Index;
        }
    }

private:
    PlayQueue&  m_PlayQueue;
    int32_t     m_Index;
};


PlayQueue::PlayQueue(MusicLibrary& library)
: m_Library(library)
{
}

PlayQueue::~PlayQueue()
{
}

void PlayQueue::loadQueue()
{
    string queuePath = determinePlayQueuePath();

    ifstream file(queuePath.c_str(), ios_base::binary);
    if (!file.is_open())
    {
        return;
    }

    file.seekg(0, ios_base::end);
    uint32_t numTracks = file.tellg() / sizeof(uint32_t);
    file.seekg(0, ios_base::beg);

    vector<uint32_t> ids(numTracks);
    file.read(reinterpret_cast<char*>(&ids[0]), numTracks * sizeof(uint32_t));

    for (size_t i = 0; i < ids.size(); ++i)
    {
        queueTrack(ids[i]);
    }

    file.close();

    FileOperations::deleteFile(queuePath);
}

void PlayQueue::saveQueue()
{
    if (m_Tracks.empty())
    {
        return;
    }

    ofstream file(determinePlayQueuePath().c_str(), ios_base::binary);
    if (!file.is_open())
    {
        return;
    }

    for (list<Track>::iterator iter = m_Tracks.begin(); iter != m_Tracks.end(); ++iter)
    {
        file.write(reinterpret_cast<char*>(&(iter->id)), sizeof(uint32_t));
    }
}

void PlayQueue::queueTrack(const Track& track, int32_t index)
{
    if (index == -1)
    {
        index = m_Tracks.size();
        m_Tracks.push_back(track);
    }
    else
    {
        list<Track>::iterator iter = m_Tracks.begin();
        for (int32_t i = 0; i < index && iter != m_Tracks.end(); ++i)
        {
            ++iter;
        }

        if (iter == m_Tracks.end())
        {
            index = m_Tracks.size();
        }

        m_Tracks.insert(iter, track);
    }

    notifyTrackQueued(index);
}

void PlayQueue::queueTrack(uint32_t id, int32_t index)
{
    try
    {
        Track track;
        m_Library.getTrack(id, track);
        queueTrack(track, index);
    }
    catch (logic_error& e)
    {
        Log::error("Failed to queue track:", e.what());
    }
}

void PlayQueue::queueAlbum(uint32_t id, int32_t index)
{
    try
    {
        QueueTrackSubscriber subscriber(*this, index);
        m_Library.getTracksFromAlbum(id, subscriber);
    }
    catch (logic_error& e)
    {
        Log::error("Failed to queue album:", e.what());
    }
}

void PlayQueue::queueRandomTracks(uint32_t count)
{
    try
    {
        QueueTrackSubscriber subscriber(*this, -1);
        m_Library.getRandomTracks(count, subscriber);
    }
    catch (logic_error& e)
    {
        Log::error("Failed to queue random tracks:", e.what());
    }
}

void PlayQueue::queueRandomAlbum()
{
    try
    {
        QueueTrackSubscriber subscriber(*this, -1);
        m_Library.getRandomAlbum(subscriber);
    }
    catch (logic_error& e)
    {
        Log::error("Failed to queue random album:", e.what());
    }
}

bool PlayQueue::getNextTrack(Track& track)
{
    if (m_Tracks.empty())
    {
        return false;
    }

    track = m_Tracks.front();
    m_Tracks.pop_front();
    notifyTrackRemoved(0);
    return true;
}

void PlayQueue::removeTrack(uint32_t index)
{
    if (m_Tracks.empty())
    {
        return;
    }

    list<Track>::iterator iter = m_Tracks.begin();
    for (uint32_t i = 0; i < index && iter != m_Tracks.end(); ++i)
    {
        ++iter;
    }

    if (iter == m_Tracks.end())
    {
        Log::error("PlayQueue::removeTrack invalid index:", index);
        return;
    }

    m_Tracks.erase(iter);
    notifyTrackRemoved(index);
}

void PlayQueue::removeTracks(std::vector<uint32_t> indexes)
{
    sort(indexes.begin(), indexes.end());

    for (size_t i = 0; i < indexes.size(); ++i)
    {
        removeTrack(indexes[i] - i);
    }
}

void PlayQueue::moveTrack(uint32_t sourceIndex, uint32_t destIndex)
{
    if (sourceIndex >= m_Tracks.size())
    {
        Log::error("PlayQueue::moveTrack: invalid fromIndex");
        return;
    }

    list<Track>::iterator sourceIter = m_Tracks.begin();
    for (uint32_t i = 0; i < sourceIndex; ++i)
    {
        ++sourceIter;
    }

    Track track = *sourceIter;
    m_Tracks.erase(sourceIter);

    list<Track>::iterator destIter = m_Tracks.begin();
    for (uint32_t i = 0; i < destIndex && destIter != m_Tracks.end(); ++i)
    {
        ++destIter;
    }

    if (destIter == m_Tracks.end())
    {
        destIndex = m_Tracks.size();
    }

    m_Tracks.insert(destIter, track);

    notifyTrackMoved(sourceIndex, destIndex);
}

bool PlayQueue::getTrackInfo(uint32_t index, Track& track)
{
    if (index >= m_Tracks.size())
    {
        return false;
    }

    list<Track>::iterator iter = m_Tracks.begin();
    for (uint32_t i = 0; i < index; ++i)
    {
        ++iter;
    }

    track = *iter;
    return true;
}

void PlayQueue::clear()
{
    m_Tracks.clear();
    notifyQueueCleared();
}

uint32_t PlayQueue::size() const
{
    return m_Tracks.size();
}

const std::list<Track>& PlayQueue::getTracks()
{
    return m_Tracks;
}

void PlayQueue::subscribe(PlayQueueSubscriber& subscriber)
{
    m_Subscribers.push_back(&subscriber);
}

void PlayQueue::unsubscribe(PlayQueueSubscriber& subscriber)
{
    for (vector<PlayQueueSubscriber*>::iterator iter = m_Subscribers.begin(); iter != m_Subscribers.end(); ++iter)
    {
        if (*iter == &subscriber)
        {
            iter = m_Subscribers.erase(iter);
            break;
        }
    }
}

void PlayQueue::notifyTrackQueued(uint32_t index)
{
    for (size_t j = 0; j < m_Subscribers.size(); ++j)
    {
        m_Subscribers[j]->onTrackQueued(index);
    }
}

void PlayQueue::notifyTrackRemoved(uint32_t index)
{
    for (size_t j = 0; j < m_Subscribers.size(); ++j)
    {
        m_Subscribers[j]->onTrackRemoved(index);
    }
}

void PlayQueue::notifyTrackMoved(uint32_t sourceIndex, uint32_t destIndex)
{
    for (size_t j = 0; j < m_Subscribers.size(); ++j)
    {
        m_Subscribers[j]->onTrackMoved(sourceIndex, destIndex);
    }
}

void PlayQueue::notifyQueueCleared()
{
    for (size_t j = 0; j < m_Subscribers.size(); ++j)
    {
        m_Subscribers[j]->onQueueCleared();
    }
}

string PlayQueue::determinePlayQueuePath()
{
    string queuePath;
    string home = FileOperations::getHomeDirectory();
    string configDir = FileOperations::combinePath(home, CONFIG_DIR);

    if (!FileOperations::pathExists(configDir))
    {
        FileOperations::createDirectory(configDir);
    }

    FileOperations::combinePath(configDir, QUEUE_FILE, queuePath);

    return queuePath;
}


}
