//    Copyright (C) 2009 Dirk Vanden Boer <dirk.vdb@gmail.com>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "librarychangedispatcher.h"

#include "Utils/scopedlock.h"
#include "MusicLibrary/track.h"
#include "MusicLibrary/album.h"

namespace Gejengel
{

LibraryChangeDispatcher::LibraryChangeDispatcher()
: m_NumTracks(0)
, m_ScannedTracks(0)
{
    sendTrackAdded.connect(sigc::mem_fun(this, &LibraryChangeDispatcher::dispatchNewTrack));
    sendTrackDeleted.connect(sigc::mem_fun(this, &LibraryChangeDispatcher::dispatchDeletedTrack));
    sendTrackUpdated.connect(sigc::mem_fun(this, &LibraryChangeDispatcher::dispatchUpdatedTrack));
    sendAlbumAdded.connect(sigc::mem_fun(this, &LibraryChangeDispatcher::dispatchNewAlbum));
    sendAlbumDeleted.connect(sigc::mem_fun(this, &LibraryChangeDispatcher::dispatchDeletedAlbum));
    sendAlbumUpdated.connect(sigc::mem_fun(this, &LibraryChangeDispatcher::dispatchUpdatedAlbum));
    sendScanStart.connect(sigc::mem_fun(this, &LibraryChangeDispatcher::dispatchScanStart));
    sendScanUpdate.connect(sigc::mem_fun(this, &LibraryChangeDispatcher::dispatchScanUpdate));
    sendScanFinish.connect(sigc::mem_fun(this, &LibraryChangeDispatcher::dispatchScanFinish));
    sendScanFailed.connect(sigc::mem_fun(this, &LibraryChangeDispatcher::dispatchScanFailed));
    sendLibraryCleared.connect(sigc::mem_fun(this, &LibraryChangeDispatcher::dispatchLibraryCleared));
}

void LibraryChangeDispatcher::addSubscriber(ILibrarySubscriber& subscriber)
{
    m_Subscribers.push_back(&subscriber);
}

void LibraryChangeDispatcher::newTrack(const Track& track)
{
    utils::ScopedLock lock(m_VectorMutex);
    m_NewTracks.push_back(track);
    sendTrackAdded();
}

void LibraryChangeDispatcher::newAlbum(const Album& album)
{
    utils::ScopedLock lock(m_VectorMutex);
    m_NewAlbums.push_back(album);
    sendAlbumAdded();
}

void LibraryChangeDispatcher::deletedTrack(uint32_t trackId)
{
    utils::ScopedLock lock(m_VectorMutex);
    m_DeletedTracks.push_back(trackId);
    sendTrackDeleted();
}

void LibraryChangeDispatcher::deletedAlbum(uint32_t albumId)
{
    utils::ScopedLock lock(m_VectorMutex);
    m_DeletedAlbums.push_back(albumId);
    sendAlbumDeleted();
}

void LibraryChangeDispatcher::updatedTrack(const Track& track)
{
    utils::ScopedLock lock(m_VectorMutex);
    m_UpdatedTracks.push_back(track);
    sendTrackUpdated();
}

void LibraryChangeDispatcher::updatedAlbum(const Album& album)
{
    utils::ScopedLock lock(m_VectorMutex);
    m_UpdatedAlbums.push_back(album);
    sendAlbumUpdated();
}

void LibraryChangeDispatcher::scanStart(uint32_t numTracks)
{
    m_NumTracks = numTracks;
    sendScanStart();
}

void LibraryChangeDispatcher::scanUpdate(uint32_t scannedTracks)
{
    m_ScannedTracks = scannedTracks;
    sendScanUpdate();
}

void LibraryChangeDispatcher::scanFinish()
{
    sendScanFinish();
}

void LibraryChangeDispatcher::scanFailed()
{
    sendScanFailed();
}

void LibraryChangeDispatcher::libraryCleared()
{
    sendLibraryCleared();
}

void LibraryChangeDispatcher::dispatchNewTrack()
{
    utils::ScopedLock lock(m_VectorMutex);
    for (size_t i = 0; i < m_NewTracks.size(); ++i)
    {
        for (size_t j = 0; j < m_Subscribers.size(); ++j)
        {
            m_Subscribers[j]->newTrack(m_NewTracks[i]);
        }
    }
    m_NewTracks.clear();
}

void LibraryChangeDispatcher::dispatchNewAlbum()
{
    utils::ScopedLock lock(m_VectorMutex);
    for (size_t i = 0; i < m_NewAlbums.size(); ++i)
    {
        for (size_t j = 0; j < m_Subscribers.size(); ++j)
        {
            m_Subscribers[j]->newAlbum(m_NewAlbums[i]);
        }
    }
    m_NewAlbums.clear();
}

void LibraryChangeDispatcher::dispatchDeletedTrack()
{
    utils::ScopedLock lock(m_VectorMutex);
    for (size_t i = 0; i < m_DeletedTracks.size(); ++i)
    {
        for (size_t j = 0; j < m_Subscribers.size(); ++j)
        {
            m_Subscribers[j]->deletedTrack(m_DeletedTracks[i]);
        }
    }
    m_DeletedTracks.clear();
}

void LibraryChangeDispatcher::dispatchDeletedAlbum()
{
    utils::ScopedLock lock(m_VectorMutex);
    for (size_t i = 0; i < m_DeletedAlbums.size(); ++i)
    {
        for (size_t j = 0; j < m_Subscribers.size(); ++j)
        {
            m_Subscribers[j]->deletedAlbum(m_DeletedAlbums[i]);
        }
    }
    m_DeletedAlbums.clear();
}

void LibraryChangeDispatcher::dispatchUpdatedTrack()
{
    utils::ScopedLock lock(m_VectorMutex);
    for (size_t i = 0; i < m_UpdatedTracks.size(); ++i)
    {
        for (size_t j = 0; j < m_Subscribers.size(); ++j)
        {
            m_Subscribers[j]->updatedTrack(m_UpdatedTracks[i]);
        }
    }
    m_UpdatedTracks.clear();
}

void LibraryChangeDispatcher::dispatchUpdatedAlbum()
{
    utils::ScopedLock lock(m_VectorMutex);
    for (size_t i = 0; i < m_UpdatedAlbums.size(); ++i)
    {
        for (size_t j = 0; j < m_Subscribers.size(); ++j)
        {
            m_Subscribers[j]->updatedAlbum(m_UpdatedAlbums[i]);
        }
    }
    m_UpdatedAlbums.clear();
}

void LibraryChangeDispatcher::dispatchScanStart()
{
    for (size_t j = 0; j < m_Subscribers.size(); ++j)
    {
        m_Subscribers[j]->scanStart(m_NumTracks);
    }
}

void LibraryChangeDispatcher::dispatchScanUpdate()
{
    for (size_t j = 0; j < m_Subscribers.size(); ++j)
    {
        m_Subscribers[j]->scanUpdate(m_ScannedTracks);
    }
}

void LibraryChangeDispatcher::dispatchScanFinish()
{
    for (size_t j = 0; j < m_Subscribers.size(); ++j)
    {
        m_Subscribers[j]->scanFinish();
    }
}

void LibraryChangeDispatcher::dispatchScanFailed()
{
    for (size_t j = 0; j < m_Subscribers.size(); ++j)
    {
        m_Subscribers[j]->scanFailed();
    }
}

void LibraryChangeDispatcher::dispatchLibraryCleared()
{
    for (size_t j = 0; j < m_Subscribers.size(); ++j)
    {
        m_Subscribers[j]->libraryCleared();
    }
}

}
