//    Copyright (C) 2009 Dirk Vanden Boer <dirk.vdb@gmail.com>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "scanner.h"

#include <cassert>
#include <ctime>

#include "track.h"
#include "album.h"
#include "musicdb.h"
#include "Utils/stringoperations.h"
#include "Utils/log.h"
#include "metadata.h"
#include "subscribers.h"
#include "Core/commonstrings.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


using namespace std;
using namespace StringOperations;

namespace Gejengel
{

Scanner::Scanner(MusicDb& db, ILibrarySubscriber& subscriber, const std::string& albumArtFilenames)
: m_LibraryDb(db)
, m_LibrarySubscriber(subscriber)
, m_ScannedFiles(0)
, m_InitialScan(false)
{
    vector<string> filenames = tokenize(albumArtFilenames, ";");
    for (size_t i = 0; i < filenames.size(); ++i)
    {
        trim(filenames[i]);
        if (!filenames[i].empty())
        {
            m_AlbumArtFilenames.push_back(filenames[i]);
        }
    }
}

void Scanner::performScan(const string& libraryPath)
{
#ifdef ENABLE_DEBUG
    time_t startTime = time(NULL);
    Log::debug("Starting library scan in:", libraryPath);
#endif

    m_InitialScan = m_LibraryDb.getTrackCount() == 0;
    m_ScannedFiles = 0;

    m_LibrarySubscriber.scanStart(FileOperations::countFilesInDirectory(libraryPath));
    FileOperations::iterateDirectory(libraryPath, *this);
    m_LibrarySubscriber.scanFinish();

#ifdef ENABLE_DEBUG
    Log::debug("Library scan took", time(NULL) - startTime, " seconds. Scanned", m_ScannedFiles, "files.");
#endif
}

bool Scanner::onFile(const string& filepath)
{
    Track track;
    track.filepath = filepath;
    FileOperations::getFileInfo(filepath, track.fileSize, track.modifiedTime);

    m_LibrarySubscriber.scanUpdate(++m_ScannedFiles);

    MusicDb::TrackStatus status = m_LibraryDb.getTrackStatus(filepath, track.modifiedTime);
    if (status == MusicDb::UpToDate)
    {
        return true;
    }

    Metadata md(track.filepath);
    if (!md.isValid())
    {
        return true;
    }

    md.getArtist(track.artist);
    md.getAlbumArtist(track.albumArtist);
    md.getTitle(track.title);
    md.getAlbum(track.album);
    md.getGenre(track.genre);
    md.getComposer(track.composer);

    track.year          = md.getYear();
    track.trackNr       = md.getTrackNr();
    track.discNr        = md.getDiscNr();
    track.bitrate       = md.getBitRate();
    track.sampleRate    = md.getSampleRate();
    track.channels      = md.getChannels();
    track.durationInSec = md.getDuration();

    if (track.album.empty())    track.album = UNKNOWN_ALBUM;
    if (track.artist.empty())   track.artist = UNKNOWN_ARTIST;
    if (track.title.empty())    track.title = UNKNOWN_TITLE;

    Album album;
    uint32_t albumId = m_LibraryDb.albumExists(track.album);
    if (albumId != 0)
    {
        if (!m_LibraryDb.getAlbum(albumId, album))
        {
            assert(false && "The corresponding album could not be found in the database");
        }
    }
    
    if (albumId == 0 || ((!track.albumArtist.empty()) && track.albumArtist != album.artist))
    {
        album.title = track.album;
        album.artist = track.albumArtist.empty() ? track.artist : track.albumArtist;
        album.year = track.year;
        album.durationInSec = track.durationInSec;
        album.dateAdded = m_InitialScan ? track.modifiedTime : time(NULL);

        md.getAlbumArt(album.coverData, m_AlbumArtFilenames);
        m_LibraryDb.addAlbum(album);
    }
    else
    {
        if (!track.albumArtist.empty())
        {
            //if a track has an album artist set, we trust it is the right one
            if (track.albumArtist != album.artist)
            {
                album.artist = track.albumArtist;
            }
        }
        else
        {
            //if no album artist set and we detect different artists for an
            //album, we set the album artist to VARIOUS_ARTISTS
            if (album.artist != VARIOUS_ARTISTS && album.artist != track.artist)
            {
                album.artist = VARIOUS_ARTISTS;
            }
        }

        if (!m_LibraryDb.getAlbumArt(albumId, album.coverData) || status == MusicDb::NeedsUpdate)
        {
            if (md.getAlbumArt(album.coverData, m_AlbumArtFilenames))
            {
                m_LibraryDb.setAlbumArt(albumId, album.coverData);
            }
        }

        if (status == MusicDb::DoesntExist)
        {
            album.durationInSec += track.durationInSec;
        }

        m_LibraryDb.updateAlbum(album);
    }

    track.albumId = album.id;
    if (status == MusicDb::DoesntExist)
    {
        Log::debug("New track:", filepath);
        m_LibraryDb.addTrack(track);
    }
    else if (status == MusicDb::NeedsUpdate)
    {
        Log::debug("Needs update:", filepath);
        m_LibraryDb.updateTrack(track);
    }

    return true;
}

}
