//    Copyright (C) 2009 Dirk Vanden Boer <dirk.vdb@gmail.com>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "albummodel.h"

#include <sstream>
#include <iomanip>

#include <glibmm/i18n.h>
#include <gdkmm/pixbufloader.h>

#include "sharedfunctions.h"
#include "Utils/log.h"
#include "Utils/trace.h"
#include "MusicLibrary/album.h"
#include "Core/albumartprovider.h"
#include "Core/commonstrings.h"

using namespace std;
using namespace Gtk;

namespace Gejengel
{

const uint32_t AlbumModel::AlbumArtNeedsUpdate = 0xFFFFFFFF;

AlbumModel::AlbumModel(IAlbumArtProvider& artProvider)
: m_AlbumArtProvider(artProvider)
{
    Utils::trace("Create Album model");
    m_ListStore = Gtk::ListStore::create(m_Columns);
}


AlbumModel::~AlbumModel()
{

}

const AlbumModel::Columns& AlbumModel::columns()
{
    return m_Columns;
}

void AlbumModel::setSortColumn(int32_t id, Gtk::SortType type)
{
    m_ListStore->set_sort_column(id, type);
}

void AlbumModel::getSortColumn(int32_t& id, Gtk::SortType& type)
{
    m_ListStore->get_sort_column_id(id, type);
}

void AlbumModel::clear()
{
    Gtk::TreeModel::Children rows = m_ListStore->children();
    while (rows.begin() != rows.end())
    {
        m_ListStore->erase(rows.begin());
    }

    //clear seams to screw up the entire model
    //m_ListStore.clear();
}

void AlbumModel::addAlbum(const Album& album)
{
    Glib::ustring duration;
    Shared::durationToString(album.durationInSec, duration);

    Gtk::TreeModel::Row row = *(m_ListStore->append());
    row[m_Columns.id]       = album.id;
    row[m_Columns.title]    = album.title == UNKNOWN_ALBUM ? _(UNKNOWN_ALBUM) : album.title;
    row[m_Columns.year]     = album.year;
    row[m_Columns.duration] = duration;
    row[m_Columns.dateAdded]= album.dateAdded;
    row[m_Columns.albumArtSize] = AlbumArtNeedsUpdate;

    if (album.artist == UNKNOWN_ARTIST)
    {
        row[m_Columns.artist] = _(UNKNOWN_ARTIST);
    }
    else if (album.artist == VARIOUS_ARTISTS)
    {
        row[m_Columns.artist] = _(VARIOUS_ARTISTS);
    }
    else
    {
        row[m_Columns.artist] = album.artist;
    }
}

void AlbumModel::deleteAlbum(uint32_t id)
{
    Gtk::TreeModel::Children rows = m_ListStore->children();

    for (Gtk::TreeModel::iterator iter = rows.begin(); iter != rows.end(); ++iter)
    {
        if ((*iter)[m_Columns.id] == id)
        {
            m_ListStore->erase(iter);
            break;
        }
    }
}

void AlbumModel::updateAlbum(const Album& album)
{
    Gtk::TreeModel::Children rows = m_ListStore->children();

    for (Gtk::TreeModel::iterator iter = rows.begin(); iter != rows.end(); ++iter)
    {
        if ((*iter)[m_Columns.id] == album.id)
        {
            Glib::ustring duration;
            Shared::durationToString(album.durationInSec, duration);

            (*iter)[m_Columns.title]    = album.title == UNKNOWN_ALBUM ? _(UNKNOWN_ALBUM) : album.title;
            (*iter)[m_Columns.year]     = album.year;
            (*iter)[m_Columns.duration] = duration;
            (*iter)[m_Columns.dateAdded]= album.dateAdded;

            if (album.coverData.size() != (*iter)[m_Columns.albumArtSize])
            {
                (*iter)[m_Columns.albumArt] = Glib::RefPtr<Gdk::Pixbuf>();
                (*iter)[m_Columns.albumArtSize] = AlbumArtNeedsUpdate;
            }

            if (album.artist == UNKNOWN_ARTIST)
            {
                (*iter)[m_Columns.artist] = _(UNKNOWN_ARTIST);
            }
            else if (album.artist == VARIOUS_ARTISTS)
            {
                (*iter)[m_Columns.artist] = _(VARIOUS_ARTISTS);
            }
            else
            {
                (*iter)[m_Columns.artist] = album.artist;
            }

            break;
        }
    }
}

bool AlbumModel::fetchAlbumArt(uint32_t albumId, int32_t size)
{
    Gtk::TreeModel::Children rows = m_ListStore->children();

    for (Gtk::TreeModel::iterator iter = rows.begin(); iter != rows.end(); ++iter)
    {
        if ((*iter)[m_Columns.id] == albumId)
        {
            Album album;
            if (m_AlbumArtProvider.getAlbumArt(albumId, album.coverData))
            {
                try
                {
                    (*iter)[m_Columns.albumArt] = Shared::createCoverPixBuf(album, size);
                    (*iter)[m_Columns.albumArtSize] = album.coverData.size();
                    return true;
                }
                catch (Gdk::PixbufError& e)
                {
                    Log::error("Failed to load album art for", albumId, "(", e.what(), ")");
                    (*iter)[m_Columns.albumArtSize] = 0;
                }
            }
            else
            {
                (*iter)[m_Columns.albumArtSize] = 0;
                return false;
            }
        }
    }

    return false;
}

void AlbumModel::clearAlbumArtCache()
{
    Gtk::TreeModel::Children rows = m_ListStore->children();
    for (Gtk::TreeModel::iterator iter = rows.begin(); iter != rows.end(); ++iter)
    {
        (*iter)[m_Columns.albumArt] = Glib::RefPtr<Gdk::Pixbuf>();
    }
}

void AlbumModel::onAlbum(const Album& album)
{
    addAlbum(album);
}

}
