//    Copyright (C) 2009 Dirk Vanden Boer <dirk.vdb@gmail.com>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "cellrendereralbum.h"

#include "Utils/log.h"
#include "Utils/trace.h"
#include "Utils/numericoperations.h"
#include "MusicLibrary/album.h"
#include "albummodel.h"
#include "mouseawaretreeview.h"

#include <assert.h>

using namespace std;
using namespace Gtk;
using namespace Gdk;
using namespace Glib;
using namespace NumericOperations;

namespace Gejengel
{

extern Gtk::Button* pButton; //OMG please fix this!

CellRendererAlbum::CellRendererAlbum(AlbumModel& model, MouseInfo& mousePos)
: Glib::ObjectBase(typeid(CellRendererAlbum))
, m_IdProperty(*this, "id", 0)
, m_AlbumArtSizeProperty(*this, "album_art_size", true)
, m_AlbumArtProperty(*this, "album", Glib::RefPtr<Gdk::Pixbuf>())
, m_TitleProperty(*this, "title", "")
, m_ArtistProperty(*this, "artist", "")
, m_YearProperty(*this, "year", 0)
, m_DurationProperty(*this, "duration", "")
, m_AlbumModel(model)
, m_MouseInfo(mousePos)
, m_Size(Large)
, m_IconSize(0)
, m_HoverSize(0)
, m_OverlaySize(0)
, m_HoverOffset(0)
, m_OverlayOffset(0)
{
    Utils::trace("Create Album cellrenderer");
    
    setSize(m_Size);

    Glib::RefPtr<IconTheme> theme = IconTheme::get_default();
    theme->signal_changed().connect(sigc::mem_fun(*this, &CellRendererAlbum::iconThemeChanged));

    property_xalign() = 0.0;
    property_yalign() = 0.5;

    Utils::trace("Album cellrenderer created");
}

CellRendererAlbum::~CellRendererAlbum()
{
}

void CellRendererAlbum::setSize(Size size)
{
    switch (size)
    {
    case Large:
        m_IconSize = 64;
        m_HoverSize = 64;
        m_OverlaySize = 32;
        m_HoverOffset = 0;
        property_xpad() = 5;
        property_ypad() = 5;
        break;
    case Small:
        m_IconSize = 32;
        m_HoverSize = 32;
        m_OverlaySize = 16;
        m_HoverOffset = 0;
        property_xpad() = 2;
        property_ypad() = 2;
        break;
    default:
        return;
    }

    m_OverlayOffset = (m_HoverSize - m_OverlaySize) / 2;
    m_Size = size;
    m_AlbumModel.clearAlbumArtCache();

    //clear images so they will be reloaded again on the next paint
    m_DefaultAlbumArt = Glib::RefPtr<Gdk::Pixbuf>();
    m_PlayImage = Glib::RefPtr<Gdk::Pixbuf>();
}

CellRendererAlbum::Size CellRendererAlbum::getSize()
{
    return m_Size;
}

void CellRendererAlbum::loadIcons()
{
    Glib::RefPtr<IconTheme> theme = IconTheme::get_default();
    m_DefaultAlbumArt = theme->load_icon("audio-x-generic", m_IconSize, (IconLookupFlags) 0);
    m_PlayImage = theme->load_icon("list-add", m_OverlaySize, (IconLookupFlags) 0);
}

Glib::PropertyProxy<uint32_t> CellRendererAlbum::property_id()
{
    return m_IdProperty.get_proxy();
}

Glib::PropertyProxy<uint32_t> CellRendererAlbum::property_album_art_size()
{
    return m_AlbumArtSizeProperty.get_proxy();
}

Glib::PropertyProxy<Glib::RefPtr<Gdk::Pixbuf> > CellRendererAlbum::property_album_art()
{
    return m_AlbumArtProperty.get_proxy();
}

Glib::PropertyProxy<Glib::ustring> CellRendererAlbum::property_title()
{
    return m_TitleProperty.get_proxy();
}

Glib::PropertyProxy<Glib::ustring> CellRendererAlbum::property_artist()
{
    return m_ArtistProperty.get_proxy();
}

Glib::PropertyProxy<uint32_t> CellRendererAlbum::property_year()
{
    return m_YearProperty.get_proxy();
}

Glib::PropertyProxy<Glib::ustring> CellRendererAlbum::property_duration()
{
    return m_DurationProperty.get_proxy();
}

static bool positionInRectangle(const MouseInfo& pos, const Rectangle& rect)
{
    int32_t rectX = rect.get_x();
    int32_t rectY = rect.get_y();
    return ((pos.x >= rectX) && (pos.x <= rectX + rect.get_width())
        &&  (pos.y >= rectY) && (pos.y <= rectY + rect.get_height()));
}

Glib::RefPtr<Pango::Layout> CellRendererAlbum::createLayout(Gtk::Widget& widget) const
{
    stringstream ss;
    ss << "<b>" << Glib::Markup::escape_text(m_TitleProperty.get_value()) << "</b>\n<i>" << Glib::Markup::escape_text(m_ArtistProperty.get_value()) << "</i>";

    if (m_Size == Large)
    {
        ss << "\n<small>";
        if (m_YearProperty.get_value() != 0)
        {
            ss << toString(m_YearProperty.get_value()) << "\n";
        }
        ss << m_DurationProperty.get_value() << "</small>";
    }

    RefPtr<Pango::Layout> layout = widget.create_pango_layout("");
    layout->set_markup(ss.str());

    return layout;
}

void CellRendererAlbum::get_size_vfunc(Widget& widget, const Rectangle* cellArea, int32_t* xOffset, int32_t* yOffset, int32_t* width, int32_t* height) const
{
    RefPtr<Pango::Layout> layout = createLayout(widget);
    Pango::Rectangle rect = layout->get_pixel_logical_extents();

    //setting the width to the actual size of the contents causes that the mainwindow
    //can not be made smaller than this value, which can be annoying when there is big text
    //the width seems to act like a minimum value 
    *width = 50;
    //*width = (property_xpad() * 3) + m_IconSize + rect.get_width();
    *height = (property_ypad() * 2) + max(m_IconSize, rect.get_height());

    if (cellArea)
    {
        if (xOffset)
        {
            *xOffset = 0;
        }

        if (yOffset)
        {
            *yOffset = max(0, int(property_yalign() * (cellArea->get_height() - rect.get_height())));
        }
    }
}

void CellRendererAlbum::render_vfunc(const RefPtr<Drawable>& window, Widget& widget, const Rectangle& bgArea, const Rectangle& cellArea, const Rectangle& exposeArea, CellRendererState flags)
{
    const uint32_t xPad = property_xpad();
    const uint32_t yPad = property_ypad();

    int32_t albumYOffset = yPad;
    RefPtr<Pango::Layout> layout = createLayout(widget);
    Pango::Rectangle rect = layout->get_pixel_logical_extents();
    if (rect.get_height() > m_IconSize)
    {
        albumYOffset += (rect.get_height() - m_IconSize) / 2;
    }

    Rectangle albumRect(cellArea.get_x() + xPad, cellArea.get_y() + albumYOffset, m_IconSize, m_IconSize);
    bool mouseInAlbumArt = positionInRectangle(m_MouseInfo, albumRect);

    if (positionInRectangle(m_MouseInfo, cellArea))
    {
        m_MouseInfo.isOnButton = mouseInAlbumArt;
        m_MouseInfo.isInRegion = mouseInAlbumArt;
    }

    if (m_MouseInfo.hasMoved && !mouseInAlbumArt)
    {
        m_MouseInfo.hasMoved = false;
        return;
    }

    int32_t xOffset = 0, yOffset = 0, width = 0, height = 0;
    get_size(widget, cellArea, xOffset, yOffset, width, height);

    width = cellArea.get_width();

    width  -= xPad * 2;
    height -= yPad * 2;

    if (width <= 0 || height <= 0)
    {
        return;
    }

    layout->set_width((widget.get_width() - m_IconSize - 4*xPad) * Pango::SCALE);
    layout->set_ellipsize(Pango::ELLIPSIZE_END);

    //Log::debug("xoffset:", xOffset, "yOffset", yOffset);
    //Log::debug("xpad:", xPad, "yPad", yPad);
    //Log::debug("Width:", width, layout->get_width());
    //Log::debug("cellw:", cellArea.get_width(), "cellh", cellArea.get_height());

    StateType textState;

    if ((flags & CELL_RENDERER_SELECTED) != 0)
    {
        textState = (widget.has_focus()) ? STATE_SELECTED : STATE_ACTIVE;
    }
    else if ((flags & CELL_RENDERER_PRELIT) != 0)
    {
        textState = STATE_PRELIGHT;
    }
    else
    {
        textState = (widget.is_sensitive()) ? STATE_NORMAL : STATE_INSENSITIVE;
    }

    RefPtr<GC> gc = widget.get_style()->get_fg_gc(widget.get_state());
    window->draw_layout(gc, cellArea.get_x() + 3 * xPad + m_IconSize + xOffset, cellArea.get_y() + yOffset, layout);

    if (m_AlbumArtSizeProperty.get_value() == AlbumModel::AlbumArtNeedsUpdate)
    {
        m_AlbumModel.fetchAlbumArt(m_IdProperty.get_value(), m_IconSize);
    }

    if (!mouseInAlbumArt || (m_HoverSize != m_IconSize))
    {
        if ((m_AlbumArtSizeProperty.get_value() != 0) && m_AlbumArtProperty.get_value())
        {
            window->draw_pixbuf(gc, m_AlbumArtProperty.get_value(), 0, 0, cellArea.get_x() + xPad, cellArea.get_y() + albumYOffset, m_IconSize, m_IconSize, RGB_DITHER_NORMAL, 0, 0);
            window->draw_rectangle(gc, false, cellArea.get_x() + xPad, cellArea.get_y() + albumYOffset, m_IconSize, m_IconSize);
        }
        else
        {
            if (!m_DefaultAlbumArt)
            {
                loadIcons();
            }
            window->draw_pixbuf(gc, m_DefaultAlbumArt, 0, 0, cellArea.get_x() + xPad, cellArea.get_y() + albumYOffset, -1, -1, RGB_DITHER_NORMAL, 0, 0);
        }
    }

    if (mouseInAlbumArt)
    {
        RefPtr<Gdk::Window> win = RefPtr<Gdk::Window>::cast_dynamic(window);
        if (win)
        {
            StateType state = m_MouseInfo.isPressed ? STATE_ACTIVE : STATE_PRELIGHT;
            if (pButton)
            {
                pButton->get_style()->paint_box(win, state, Gtk::SHADOW_NONE, cellArea, widget, "button", cellArea.get_x() + xPad + m_HoverOffset, cellArea.get_y() + albumYOffset + m_HoverOffset, m_HoverSize, m_HoverSize);
            }
        }

        if (!m_PlayImage)
        {
            loadIcons();
        }
        window->draw_pixbuf(gc, m_PlayImage, 0, 0, cellArea.get_x() + xPad + m_HoverOffset + m_OverlayOffset, cellArea.get_y() + albumYOffset + m_HoverOffset + m_OverlayOffset, m_OverlaySize, m_OverlaySize, RGB_DITHER_NORMAL, 0, 0);
    }
}

void CellRendererAlbum::iconThemeChanged()
{
    Glib::RefPtr<IconTheme> theme = IconTheme::get_default();
    m_DefaultAlbumArt = theme->load_icon("audio-x-generic", m_IconSize, (IconLookupFlags) 0);
    m_PlayImage = theme->load_icon("media-playback-start", m_OverlaySize, (IconLookupFlags) 0);
}

void CellRendererAlbum::setButtonStyle(Glib::RefPtr<Gtk::Style> style)
{
    if (!style)
    {
        Log::debug("BAD STYLE");
    }

    m_ButtonStyle = style->copy();
}

}
