//    Copyright (C) 2009 Dirk Vanden Boer <dirk.vdb@gmail.com>
//
//    This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef WIN32
#include "winconfig.h"
#endif

#include "mainwindow.h"

#include <cassert>
#include <sstream>
#include <glibmm/i18n.h>

#include "uilayout.h"
#include "preferencesdlg.h"
#include "Core/gejengelcore.h"
#include "Core/settings.h"
#include "Core/musiclibrary.h"
#include "Core/pluginmanager.h"
#include "Utils/log.h"
#include "Utils/trace.h"

using namespace Gtk;
using namespace std;

namespace Gejengel
{

MainWindow::MainWindow(IGejengelCore& core)
: m_Core(core)
, m_AlbumModel(core.getAlbumArtProvider())
, m_PlayQueueModel(core.getPlayQueue(), core.getAlbumArtProvider())
, m_LayoutStyle(UILayoutFactory::None)
, m_pLayout(NULL)
, m_CloseToTray(true)
{
    Utils::trace("Create Mainwindow");
    set_title(PACKAGE_NAME);
    set_resizable(true);

    Glib::set_application_name(PACKAGE_NAME);
    //set_default_icon_name("cdrom");
    //Gtk::Invisible inv;
    //set_icon(inv.render_icon(Stock::CDROM, ICON_SIZE_DIALOG));

    Utils::trace("Get albums");

    try
    {
        m_Core.getMusicLibrary().getAlbums(m_AlbumModel);
    }
    catch (logic_error& e)
    {
        pushStatusMessage(string(_("Failed to get albums from database: ")) + e.what());
    }

    m_Core.getPlayQueue().subscribe(m_PlayQueueModel);

    Utils::trace("Init ui");
    init();

    loadSettings();

    add(m_EntireLayout);

    Utils::trace("Show children");
    show_all_children();

    if (m_Core.getSettings().getAsBool("ScanAtStartup", false))
    {
        scanLibrary(false);
    }

    signal_key_press_event().connect(sigc::mem_fun(*this, &MainWindow::onKeyPress), false);
    signal_window_state_event().connect(sigc::mem_fun(*this, &MainWindow::onWindowStateChanged));
    
    Utils::trace("Mainwindow created");
}

MainWindow::~MainWindow()
{
    delete m_pLayout;
    saveSettings();
}

void MainWindow::loadSettings()
{
    int32_t width = m_Core.getSettings().getAsInt("MainWindowWidth", 650);
    int32_t height = m_Core.getSettings().getAsInt("MainWindowHeight", 480);
    UILayoutFactory::LayoutStyle style = static_cast<UILayoutFactory::LayoutStyle>(m_Core.getSettings().getAsInt("LayoutStyle", UILayoutFactory::DetailedAlbums));

    //toggling the radio action causes the right layout to be loaded, the simple layout needs to be disabled first as it is active by default
    switch(style)
    {
    case UILayoutFactory::SimpleAlbums:
        Glib::RefPtr<RadioAction>::cast_dynamic(m_UIManager->get_action("/MenuBar/ViewMenu/SimpleAlbum"))->set_active(false);
        Glib::RefPtr<RadioAction>::cast_dynamic(m_UIManager->get_action("/MenuBar/ViewMenu/SimpleAlbum"))->set_active(true);
        break;
    case UILayoutFactory::Basic:
        Glib::RefPtr<RadioAction>::cast_dynamic(m_UIManager->get_action("/MenuBar/ViewMenu/Basic"))->set_active(true);
        break;
    case UILayoutFactory::DetailedAlbums:
    default:
        Glib::RefPtr<RadioAction>::cast_dynamic(m_UIManager->get_action("/MenuBar/ViewMenu/DetailedAlbum"))->set_active(true);
    }
    
    m_CloseToTray = m_Core.getSettings().getAsBool("CloseToTray", true);

    set_default_size(width, height);
}

void MainWindow::saveSettings()
{
    m_Core.getSettings().set("MainWindowWidth", get_width());
    m_Core.getSettings().set("MainWindowHeight", get_height());
    m_Core.getSettings().set("LayoutStyle", m_LayoutStyle);
}

bool MainWindow::on_delete_event(GdkEventAny* pEvent)
{
	if (m_CloseToTray)
	{
		if (!get_skip_taskbar_hint())
		{
			set_skip_taskbar_hint(true);
			iconify();
		}
	}
	else
	{
		quit();
	}
}

void MainWindow::run()
{
    Main::run(*this);
}

void MainWindow::quit()
{
    Main::quit();
}

void MainWindow::pushStatusMessage(const Glib::ustring& message)
{
    m_StatusBar.pop();
    m_StatusBar.push(message);
}

void MainWindow::showPreferences()
{
    string libraryBefore = m_Core.getSettings().get("MusicLibrary");
    bool sysTrayBefore   = m_Core.getSettings().getAsBool("TrayIcon", true);

    {
        PreferencesDlg dlg(*this, m_Core);
        dlg.run();
    }

    string libraryAfter = m_Core.getSettings().get("MusicLibrary");
    bool sysTrayAfter   = m_Core.getSettings().getAsBool("TrayIcon", true);

    if (libraryBefore != libraryAfter)
    {
        Log::info("Library location changed from", libraryBefore, "to", libraryAfter);
        scanLibrary(true);
    }
    
    if (sysTrayBefore != sysTrayAfter)
    {
        sysTrayAfter ? m_SystemTray.show() : m_SystemTray.hide();
    }
    
    m_CloseToTray = sysTrayAfter ? m_Core.getSettings().getAsBool("CloseToTray", true) : false;

    m_Core.getSettings().saveToFile();
}

void MainWindow::showAbout()
{
    Gtk::AboutDialog aboutDlg;
    aboutDlg.set_transient_for(*this);
    aboutDlg.set_name(PACKAGE_NAME);
    aboutDlg.set_version(PACKAGE_VERSION);
    aboutDlg.set_copyright("Dirk Vanden Boer");
    aboutDlg.set_license("GPLv2");
    aboutDlg.set_website("http://code.google.com/p/gejengel/");

    std::list<Glib::ustring> authors;
    authors.push_back("Dirk Vanden Boer");
    aboutDlg.set_authors(authors);

    std::list<Glib::ustring> artists;
    artists.push_back("gorvan.com (Last.fm logo)");
    aboutDlg.set_artists(artists);
    
    aboutDlg.set_translator_credits("Spanish: mati86dl");

    aboutDlg.run();
}

void MainWindow::onScanLibrary()
{
    scanLibrary(false);
}

void MainWindow::scanLibrary(bool freshStart)
{
    try
    {
        Glib::RefPtr<Action> scanAction = Glib::RefPtr<Action>::cast_dynamic(m_UIManager->get_action("/MenuBar/FileMenu/FileRescanLibrary"));
        scanAction->set_sensitive(false);

        pushStatusMessage(_("Starting library scan, obtaining directory information"));
        m_Core.getMusicLibrary().scan(freshStart);
    }
    catch (logic_error& e)
    {
        pushStatusMessage(string(_("Scan library failed: ")) + e.what());
    }
}

void MainWindow::newTrack(const Track& track)
{
    pushStatusMessage(_("Added ") + track.artist + " - " + track.title);
}

void MainWindow::newAlbum(const Album& album)
{
    m_AlbumModel.addAlbum(album);
}

void MainWindow::deletedTrack(uint32_t id)
{
    m_TrackModel.deleteTrack(id);
}

void MainWindow::deletedAlbum(uint32_t id)
{
    m_AlbumModel.deleteAlbum(id);
}

void MainWindow::updatedAlbum(const Album& album)
{
    m_AlbumModel.updateAlbum(album);
}

void MainWindow::scanStart(uint32_t numTracks)
{
    m_TracksToScan = numTracks;
    m_ScanProgress.set_fraction(0.0);
    m_ScanProgress.set_text(_("Scanning library"));
    pushStatusMessage("");
    m_InfoLayout.pack_start(m_ScanProgress, Gtk::PACK_SHRINK);
    m_ScanProgress.show();
}

void MainWindow::scanUpdate(uint32_t scannedTracks)
{
    m_ScanProgress.set_fraction(static_cast<double>(scannedTracks) / m_TracksToScan);
}

void MainWindow::scanFinish()
{
    m_InfoLayout.remove(m_ScanProgress);
    pushStatusMessage(_("Library scan finished"));

    Glib::RefPtr<Action> scanAction = Glib::RefPtr<Action>::cast_dynamic(m_UIManager->get_action("/MenuBar/FileMenu/FileRescanLibrary"));
    scanAction->set_sensitive(true);
}

void MainWindow::scanFailed()
{
    m_InfoLayout.remove(m_ScanProgress);
    pushStatusMessage(_("Library scan failed"));

    Glib::RefPtr<Action> scanAction = Glib::RefPtr<Action>::cast_dynamic(m_UIManager->get_action("/MenuBar/FileMenu/FileRescanLibrary"));
    scanAction->set_sensitive(true);
}

void MainWindow::libraryCleared()
{
    m_AlbumModel.clear();
    m_TrackModel.clear();
    m_Core.getPlayQueue().clear();
}

void MainWindow::selectedAlbumChanged(uint32_t albumId)
{
    m_TrackModel.clear();
    m_Core.getMusicLibrary().getTracksFromAlbum(albumId, m_TrackModel);
}

void MainWindow::showHideWindow()
{
    if (get_skip_taskbar_hint())
    {
        set_skip_taskbar_hint(false);
        present();
    }
    else
    {
        set_skip_taskbar_hint(true);
        iconify();
    }
}

void MainWindow::init()
{
    m_ActionGroup = Gtk::ActionGroup::create();

    m_ActionGroup->add(Action::create("FileMenu", _("_File")));
    m_ActionGroup->add(Action::create("FileRescanLibrary", Gtk::Stock::REFRESH, _("_Rescan library")), Gtk::AccelKey('r', Gdk::CONTROL_MASK), sigc::mem_fun(*this, &MainWindow::onScanLibrary));
    m_ActionGroup->add(Action::create("FileQuit", Gtk::Stock::QUIT), sigc::mem_fun(*this, &MainWindow::quit));

    m_ActionGroup->add(Action::create("EditMenu", _("_Edit")));
    m_ActionGroup->add(Action::create("EditPreferences", Gtk::Stock::PREFERENCES), Gtk::AccelKey('p', Gdk::CONTROL_MASK), sigc::mem_fun(*this, &MainWindow::showPreferences));

    m_ActionGroup->add(Action::create("ViewMenu", _("_View")));
    m_ActionGroup->add(ToggleAction::create("ViewFullscreen", Gtk::Stock::FULLSCREEN), Gtk::AccelKey("F11"), sigc::mem_fun(*this, &MainWindow::onFullscreen));

    Gtk::RadioButtonGroup layoutGroup;
    m_ActionGroup->add(RadioAction::create(layoutGroup, "SimpleAlbum", _("Simple album view")), Gtk::AccelKey("F5"), sigc::mem_fun(*this, &MainWindow::onLayoutChange));
    m_ActionGroup->add(RadioAction::create(layoutGroup, "DetailedAlbum", _("Detailed album view")), Gtk::AccelKey("F6"), sigc::mem_fun(*this, &MainWindow::onLayoutChange));
    m_ActionGroup->add(RadioAction::create(layoutGroup, "Basic", _("Basic view")), Gtk::AccelKey("F7"), sigc::mem_fun(*this, &MainWindow::onLayoutChange));

    m_ActionGroup->add(Action::create("ToolsMenu", _("_Tools")));
    m_ActionGroup->add(Action::create("ToolsRandomTracks", _("Queue random tracks")), Gtk::AccelKey('t', Gdk::CONTROL_MASK), sigc::mem_fun(*this, &MainWindow::onQueueRandomTracks));
    m_ActionGroup->add(Action::create("ToolsRandomAlbum", _("Queue random album")), Gtk::AccelKey('b', Gdk::CONTROL_MASK), sigc::mem_fun(*this, &MainWindow::onQueueRandomAlbum));
    m_ActionGroup->add(Action::create("ToolsClearQueue", _("Clear play queue")), Gtk::AccelKey('c', Gdk::CONTROL_MASK| Gdk::SHIFT_MASK), sigc::mem_fun(m_Core.getPlayQueue(), &PlayQueue::clear));

    m_ActionGroup->add(Action::create("HelpMenu", _("_Help")));
    m_ActionGroup->add(Action::create("HelpAbout", Gtk::Stock::ABOUT), sigc::mem_fun(*this, &MainWindow::showAbout));

    m_UIManager = Gtk::UIManager::create();
    m_UIManager->insert_action_group(m_ActionGroup);
    add_accel_group(m_UIManager->get_accel_group());

    Glib::ustring uiInfo =
        "<ui>"
        "  <menubar name='MenuBar'>"
        "    <menu action='FileMenu'>"
        "      <menuitem action='FileRescanLibrary'/>"
        "      <separator/>"
        "      <menuitem action='FileQuit'/>"
        "    </menu>"
        "    <menu action='EditMenu'>"
        "      <menuitem action='EditPreferences'/>"
        "    </menu>"
        "    <menu action='ViewMenu'>"
        "      <menuitem action='ViewFullscreen'/>"
        "      <menuitem action='SimpleAlbum'/>"
        "      <menuitem action='DetailedAlbum'/>"
        "      <menuitem action='Basic'/>"
        "    </menu>"
        "    <menu action='ToolsMenu'>"
        "      <menuitem action='ToolsRandomTracks'/>"
        "      <menuitem action='ToolsRandomAlbum'/>"
        "      <menuitem action='ToolsClearQueue'/>"
        "    </menu>"
        "    <menu action='HelpMenu'>"
        "      <menuitem action='HelpAbout'/>"
        "    </menu>"
        "  </menubar>"
        "</ui>";

    try
    {
        m_UIManager->add_ui_from_string(uiInfo);
    }
    catch(const Glib::Error& ex)
    {
        Log::error("Building menus failed: " +  ex.what());
    }

    Gtk::Widget* pMenubar = m_UIManager->get_widget("/MenuBar");

    m_InfoLayout.set_border_width(5);
    m_InfoLayout.set_spacing(5);

    m_EntireLayout.pack_start(*pMenubar, PACK_SHRINK);
    m_EntireLayout.pack_start(m_InfoLayout, PACK_EXPAND_WIDGET);
    m_EntireLayout.pack_start(m_StatusBar, PACK_SHRINK);
}

void MainWindow::onFullscreen()
{
    Glib::RefPtr<ToggleAction> toggle = Glib::RefPtr<ToggleAction>::cast_dynamic(m_UIManager->get_action("/MenuBar/ViewMenu/ViewFullscreen"));
    toggle->get_active() ? fullscreen() : unfullscreen();
}

void MainWindow::onQueueRandomTracks()
{
    m_Core.getPlayQueue().queueRandomTracks(10);
}

void MainWindow::onQueueRandomAlbum()
{
    m_Core.getPlayQueue().queueRandomAlbum();
}

void MainWindow::onSearchChanged(const Glib::ustring& searchString)
{
    try
    {
        m_AlbumModel.clear();
        if (searchString.empty())
        {
            m_Core.getMusicLibrary().getAlbums(m_AlbumModel);
            return;
        }

        m_TrackModel.clear();

        if (searchString.size() > 2)
        {
            m_Core.getMusicLibrary().search(searchString, m_TrackModel, m_AlbumModel);
        }
    }
    catch (exception& e)
    {
        pushStatusMessage(string(_("Failed to perform search query: ")) + e.what());
    }
}

void MainWindow::onLayoutChange()
{
    if (Glib::RefPtr<RadioAction>::cast_dynamic(m_UIManager->get_action("/MenuBar/ViewMenu/SimpleAlbum"))->get_active())
    {
        if (m_LayoutStyle != UILayoutFactory::SimpleAlbums)
        {
            m_LayoutStyle = UILayoutFactory::SimpleAlbums;
        }
        else
        {
            return;
        }
    }
    else if (Glib::RefPtr<RadioAction>::cast_dynamic(m_UIManager->get_action("/MenuBar/ViewMenu/DetailedAlbum"))->get_active())
    {
        if (m_LayoutStyle != UILayoutFactory::DetailedAlbums)
        {
            m_LayoutStyle = UILayoutFactory::DetailedAlbums;
        }
        else
        {
            return;
        }
    }
    else if (Glib::RefPtr<RadioAction>::cast_dynamic(m_UIManager->get_action("/MenuBar/ViewMenu/Basic"))->get_active())
    {
        if (m_LayoutStyle != UILayoutFactory::Basic)
        {
            m_LayoutStyle = UILayoutFactory::Basic;
        }
        else
        {
            return;
        }
    }

    loadLayout();
}

void MainWindow::loadLayout()
{
    if (m_pLayout)
    {
        m_InfoLayout.remove(m_pLayout->getWidget());
        delete m_pLayout;
    }
    
    m_pLayout = UILayoutFactory::create(m_LayoutStyle, m_Core.getSettings(), m_PlayQueueModel, m_TrackModel, m_AlbumModel);
    m_InfoLayout.pack_start(m_pLayout->getWidget(), PACK_EXPAND_WIDGET);
    show_all_children();

    vector<GejengelPlugin*> plugins;
    m_pLayout->getPlugins(plugins);
    plugins.push_back(&m_SystemTray);
    m_Core.getPluginManager().setUIPlugins(plugins);
    
    if (m_Core.getSettings().getAsBool("TrayIcon", true))
    {
        m_SystemTray.show();
    }
    else
    {
        m_SystemTray.hide();
    }

    m_pLayout->signalAlbumChanged.connect(sigc::mem_fun(*this, &MainWindow::selectedAlbumChanged));
    m_pLayout->signalSearchChanged.connect(sigc::mem_fun(*this, &MainWindow::onSearchChanged));

    onSearchChanged("");
}

bool MainWindow::onKeyPress(GdkEventKey* pEvent)
{
    switch (pEvent->keyval)
    {
    case GDK_F:
    case GDK_f:
        if (GDK_CONTROL_MASK & pEvent->state)
        {
            m_pLayout->focusSearchField();
            return true;
        }
    default:
        return false;
    }
}

bool MainWindow::onWindowStateChanged(GdkEventWindowState* pState)
{
    return false;
}

}
